/*
 * io.c
 *
 * Copyright (c) 2014 Kirill Kryukov
 *
 * This file is a part of the Nucleotide Sequence Clusterizer source code,
 * and may only be used, modified, and distributed under the terms of the
 * license, located in the file LICENSE.txt within this package.
 */

#ifdef INCLUDED_FROM_ALL_C

/*
 * Buffered file.
 */
typedef struct file_t
{
    FILE *stream;
    char *buf;
    size_t buf_size;  /* Size of buffer. */
    char *buf_start;  /* Start of unparsed data in buffer. */
    char *buf_end;    /* End of data. */
} file_t;

/*
 * Frees the file buffer.
 */
void free_file_buffer(file_t *file)
{
    assert(file);

    if (file->buf)
    {
        free1(file->buf,file->buf_size);
        file->buf = 0;
        file->buf_size = 0;
        file->buf_start = 0;
        file->buf_end = 0;
    }
}

/*
 * Allocates memory for the file buffer.
 */
void allocate_file_buffer(file_t *file,size_t buffer_size)
{
    assert(file);
    assert(buffer_size);

    free_file_buffer(file);
    file->buf = (char*)malloc1(buffer_size);
    file->buf_size = buffer_size;
    file->buf_start = file->buf;
    file->buf_end = file->buf;
}

/*
 * Closes the file stream.
 * Does not free the buffer or tfile structure.
 */
void close_file(file_t *file)
{
    assert(file);

    if (file->stream)
    {
        fclose(file->stream);
        file->stream = 0;
    }

    file->buf_start = file->buf;
    file->buf_end = file->buf;
}

/*
 * Makes a file for an already opened stream.
 * Allocates the tfile structure, if 'file' is 0.
 * Returns file address.
 */
file_t* make_buffered_file_from_stream(file_t *file,FILE *stream,size_t buffer_size)
{
    file_t *f;

    assert(stream);
    assert(buffer_size);

    f = file ? file : (file_t*)calloc1(1,sizeof(file_t));
    f->stream = stream;
    allocate_file_buffer(f,buffer_size);

    return f;
}

/*
 * Opens a file and allocates buffer.
 * If 'file' is 0, allocates the tfile structure too.
 * Returns file address on success, 0 on failure.
 */
file_t* open_file(file_t *file,char *path,char *mode,size_t buffer_size)
{
    FILE *s;

    assert(path);
    assert(mode);
    assert(buffer_size);

    if (file && file->stream) { close_file(file); }

    s = fopen(path,mode);
    if (!s) { return 0; }

    return make_buffered_file_from_stream(file,s,buffer_size);
}

/*
 * Returns 1 if the end of file is reached, 0 otherwise.
 * The file must be already opened for reading.
 */
int file_eof(file_t *file)
{
    assert(file);
    assert(file->stream);

    return (feof(file->stream) && file->buf_start >= file->buf_end);
}

/*
 * If file buffer is empty, fills the buffer.
 * Returns size of data in buffer.
 * If buffer is empry and nothing could be read from file, returns 0.
 */
ptrdiff_t read_file(file_t *file)
{
    assert(file);
    assert(file->stream);

    if (file->buf_start >= file->buf_end)
    {
        size_t bytes_read = fread(file->buf,1,file->buf_size,file->stream);
        file->buf_start = file->buf;
        file->buf_end = file->buf + bytes_read;
    }
    return file->buf_end - file->buf_start;
}

/*
 * Returns the size of the current contents of buffer.
 */
ptrdiff_t file_buffer_content_size(file_t *file)
{
    assert(file);
    return file->buf_end - file->buf_start;
}

/*
 * Reads the file and skips everything until it finds a line break.
 */
void skip_until_line_break(file_t *file)
{
    assert(file);

    while (read_file(file))
    {
        while (file->buf_start < file->buf_end && *file->buf_start!=13 && *file->buf_start!=10) { file->buf_start++; }
        if (file->buf_start < file->buf_end) { break; }
    }
}

/*
 * Reads the file and skips all line breaks, stops when any other character is found.
 */
void skip_line_breaks(file_t *file)
{
    assert(file);

    while (read_file(file))
    {
        while (file->buf_start < file->buf_end && (*file->buf_start==13 || *file->buf_start==10)) { file->buf_start++; }
        if (file->buf_start < file->buf_end) { break; }
    }
}

/*
 * Reads the file and skips everything until it finds '>'.
 */
void skip_until_next_fasta_sequence(file_t *file)
{
    assert(file);

    while (read_file(file))
    {
        while (file->buf_start < file->buf_end && *file->buf_start!='>') { file->buf_start++; }
        if (file->buf_start < file->buf_end) { break; }
    }
}

/*
 * Reads the file and skips all space characters, stops when any non-space character is found.
 */
int read_char_from_file(file_t *file)
{
    int c;
    assert(file);

    if (!read_file(file)) { return 0; }
    c = *file->buf_start;
    file->buf_start++;
    return c;
}

/*
 * Reads the file and skips all space characters, stops when any non-space character is found.
 */
void skip_spaces(file_t *file)
{
    assert(file);

    while (read_file(file))
    {
        while (file->buf_start < file->buf_end && isspace(*(unsigned char*)file->buf_start)) { file->buf_start++; }
        if (file->buf_start < file->buf_end) { break; }
    }
}

#if 0
/*
 * Reads text from file until end of line, stores in 'str'.
 * Returns the number added characters.
 */
ptrdiff_t str_add_from_file_until_line_break(tstr *str,tfile *file)
{
    size_t old_length;

    assert(str);
    assert(file);

    old_length = str->length;

    while (read_file(file))
    {
        char *c;
        for (c=file->buf_start; c<file->buf_end && *c!=13 && *c!=10; c++) {}
        if (c > file->buf_start) { store_data_in_str(str,file->buf_start,c-file->buf_start); file->buf_start = c; }
        if (c < file->buf_end) { break; }
    }

    return str->length - old_length;
}

/*
 * Reads text from file until any space character (including tabs and line breaks).
 * Returns the number of added characters.
 */
ptrdiff_t str_add_from_file_until_space(tstr *str,tfile *file)
{
    size_t old_length;

    assert(str);
    assert(file);

    old_length = str->length;

    while (read_file(file))
    {
        char *c;
        for (c=file->buf_start; c<file->buf_end && !isspace(*(unsigned char*)c); c++) {}
        if (c > file->buf_start) { store_data_in_str(str,file->buf_start,c-file->buf_start); file->buf_start = c; }
        if (c < file->buf_end) { break; }
    }

    return str->length - old_length;
}

/*
 * Reads text from file until tab or end-of-line character.
 * Returns the number of added characters.
 */
ptrdiff_t str_add_from_file_until_tab_or_eol(tstr *str,tfile *file)
{
    size_t old_length;

    assert(str);
    assert(file);

    old_length = str->length;

    while (read_file(file))
    {
        char *c;
        for (c=file->buf_start; c<file->buf_end && *c!='\t' && *c!=13 && *c!=10; c++) {}
        if (c > file->buf_start) { store_data_in_str(str,file->buf_start,c-file->buf_start); file->buf_start = c; }
        if (c < file->buf_end) { break; }
    }

    return str->length - old_length;
}
#endif

#endif
