/*
 * main.c
 *
 * Copyright (c) 2014 Kirill Kryukov
 *
 * This file is a part of the Nucleotide Sequence Clusterizer source code,
 * and may only be used, modified, and distributed under the terms of the
 * license, located in the file LICENSE.txt within this package.
 */

#ifdef INCLUDED_FROM_ALL_C

/*
 * Exit handler, deallocates memory.
 */
void done()
{
    if (template) { free1(template,template_length); template = 0; }
    if (use_barcode_position) { free1(use_barcode_position,template_length); use_barcode_position = 0; }

    while (first_block)
    {
        block_t *next = first_block->next;
        free1(first_block,sizeof(block_t));
        first_block = next;
    }

    if (index)
    {
        int i;
        for (i=0; i<N_VOLUMES; i++)
        {
            if (index[i]) { free1(index[i],volume_allocated_sizes[i]); }
        }
        free1(index,sizeof(unit_t*)*N_VOLUMES);
        index = 0;
    }
    if (volume_allocated_sizes) { free1(volume_allocated_sizes,sizeof(*volume_allocated_sizes)*N_VOLUMES); volume_allocated_sizes = 0; }
    if (volume_sizes) { free1(volume_sizes,sizeof(*volume_sizes)*N_VOLUMES); volume_sizes = 0; }

    if (cluster_entries) { free1(cluster_entries,cluster_entry_size*n_unique_barcodes); cluster_entries = 0; }
    if (clustered_barcode_index) { free1(clustered_barcode_index,sizeof(void*)*N_VOLUMES); clustered_barcode_index = 0; }

    if (allocated) { fprintf(stderr,"Error: Failed to free %lld bytes\n",allocated); }
    /*else { fprintf(stderr,"Shutdown clean\n"); }*/
}

void init_counter()
{
    int i;

    memset(is_nucleotide,0,sizeof(*is_nucleotide)*256);
    is_nucleotide['a'] = 1; is_nucleotide['A'] = 1;
    is_nucleotide['c'] = 2; is_nucleotide['C'] = 2;
    is_nucleotide['g'] = 3; is_nucleotide['G'] = 3;
    is_nucleotide['t'] = 4; is_nucleotide['T'] = 4;

    template_length = strlen(template_parameter);
    template = malloc1(template_length);

    use_barcode_position = malloc1(template_length);

    for (i=0; i<template_length; i++)
    {
        unsigned char c = ((unsigned char*)template_parameter)[i];
        unsigned char uc = toupper(c);
        if (uc == '.' || uc == 'N')
        {
            use_barcode_position[i] = 1;
            template[i] = uc;
            barcode_complexity++;
        }
        else if (is_nucleotide[uc])
        {
            use_barcode_position[i] = 0;
            template[i] = uc;
        }
        else
        {
            fprintf(stderr,"Can't parse \"%c\" in the template sequence \"%s\" - template should only contain [ATCGN.]\n",c,template);
            exit(1);
        }
    }

    /*fprintf(stderr,"Template length: %d, Complexity: %d\n",template_length,barcode_complexity);*/

    units_per_barcode = (barcode_complexity + (BP_PER_UNIT-1)) / BP_PER_UNIT;
    bytes_per_barcode = units_per_barcode * BYTES_PER_UNIT;
    /*fprintf(stderr,"Using %d units (%d bytes) to store each barcode\n",units_per_barcode,bytes_per_barcode);*/

    barcodes_per_block = BYTES_PER_BLOCK / bytes_per_barcode;
    /*fprintf(stderr,"Using blocks of %d barcodes (%d bytes)\n",barcodes_per_block,BYTES_PER_BLOCK);*/

    /*index_shift = (BP_PER_UNIT - index_depth) * 2;*/
    /*n_volumes = 1 << (index_depth * 2);*/
    /*fprintf(stderr,"index depth: %u, %d volumes, shift: %u\n",index_depth,n_volumes,index_shift);*/
    volume_allocated_sizes = (long long*)calloc1(N_VOLUMES,sizeof(long long));
    volume_sizes = (long long*)calloc1(N_VOLUMES,sizeof(long long));
    index = (unit_t**)calloc1(N_VOLUMES,sizeof(unit_t*));

    cluster_entry_size = sizeof(cluster_entry_header_t) + bytes_per_barcode;
}

void print_all_barcode_blocks()
{
    block_t *b = first_block;
    fprintf(stderr,"Printing all barcodes from input buffer\n");
    while (b)
    {
        int n = (b == last_block) ? n_barcodes_in_last_block : barcodes_per_block;
        int i,j;
        for (i=0; i<n; i++)
        {
            for (j=0; j<units_per_barcode; j++) { printf("%21llu",b->units[i*units_per_barcode+j]); }
            printf("  ");
            print_complete_barcode(&b->units[i*units_per_barcode]);
            printf("\n");
        }

        b = b->next;
    }
    fprintf(stderr," - OK\n");
}

void print_all_barcode_volumes()
{
    int i,j,k;
    for (i=0; i<N_VOLUMES; i++)
    {
        for (k=0; k<volume_sizes[i]; k++)
        {
            for (j=0; j<units_per_barcode; j++) { printf("%21llu",index[i][k*units_per_barcode+j]); }
            printf("  ");
            print_complete_barcode(&index[i][k*units_per_barcode]);
            printf("\n");
        }
    }
}

void store_barcodes_in_volumes()
{
    int i;
    long long n_stored = 0ll;

    /*fprintf(stderr,"Sorting barcodes into %d volumes\n",N_VOLUMES);*/

    block_t *b = first_block;
    while (b)
    {
        int n = (b == last_block) ? n_barcodes_in_last_block : barcodes_per_block;
        for (i=0; i<n; i++)
        {
            unsigned vol_index = b->units[i*units_per_barcode] >> INDEX_SHIFT;
            volume_allocated_sizes[vol_index]++;
        }
        b = b->next;
    }

    /*printf("Volume sizes:\n");*/
    for (i=0; i<N_VOLUMES; i++)
    {
        /*printf("    %5d - %lld",i,volume_allocated_sizes[i]);*/
        volume_allocated_sizes[i] *= bytes_per_barcode;
        /*printf(" -> %lld\n",volume_allocated_sizes[i]);*/
        if (volume_allocated_sizes[i]) { index[i] = malloc1(volume_allocated_sizes[i]); }
    }

    b = first_block;
    while (b)
    {
        int n = (b == last_block) ? n_barcodes_in_last_block : barcodes_per_block;
        for (i=0; i<n; i++)
        {
            unit_t *src = &b->units[i*units_per_barcode];
            unit_t *src_end = src + units_per_barcode;
            unsigned vol_index = (*src) >> INDEX_SHIFT;
            unit_t *dest = &index[vol_index][volume_sizes[vol_index]*units_per_barcode];
            for (; src < src_end; src++,dest++) { *dest = *src; }
            volume_sizes[vol_index]++;
            n_stored++;
        }
        b = b->next;
    }

    /*printf("Volume sizes:\n");*/
    for (i=0; i<N_VOLUMES; i++)
    {
        /*printf("    %5d - %lld - %lld\n",i,volume_allocated_sizes[i]/bytes_per_barcode,volume_sizes[i]);*/
        assert(volume_sizes[i] * bytes_per_barcode == volume_allocated_sizes[i]);
    }

    /*fprintf(stderr,"Stored %lld barcodes in indexed volumes, now %lld bytes are allocated\n",n_stored,allocated);*/

    while (first_block)
    {
        block_t *next = first_block->next;
        free1(first_block,sizeof(block_t));
        first_block = next;
    }

    /*fprintf(stderr,"Deallocated input barcode buffer, now %lld bytes are allocated\n",allocated);*/
}

void sort_barcodes()
{
    int i;
    /*fprintf(stderr,"Sorting barcodes ..");*/
    for (i=0; i<N_VOLUMES; i++)
    {
        if (index[i])
        {
            qsort(index[i],volume_sizes[i],bytes_per_barcode,&compare_barcodes);
        }
    }
    /*fprintf(stderr," OK\n");*/
}

void remove_duplicates()
{
    int i;
    fprintf(stderr,"Removing duplicates ..");
    for (i=0; i<N_VOLUMES; i++)
    {
        unit_t *a,*b,*end;

        if (volume_sizes[i] < 2) { continue; }

        a = index[i];
        b = index[i] + units_per_barcode;
        end = a + volume_sizes[i]*units_per_barcode;

        for (; b < end; b += units_per_barcode)
        {
            if (!same_barcodes(a,b))
            {
                a += units_per_barcode;
                copy_barcode(a,b);
            }
        }
        volume_sizes[i] = ((a + units_per_barcode) - index[i]) / units_per_barcode;
    }
    fprintf(stderr," OK");

    for (i=0; i<N_VOLUMES; i++)
    {
        n_unique_barcodes += volume_sizes[i];
    }
    fprintf(stderr,", %lld unique sequences remain\n",n_unique_barcodes);
}

int main(int argc,char **argv)
{
    atexit(done);

    setbuf(stderr,0);
#ifdef DEBUG
    setbuf(stdout,0);
#endif

    read_command_line(argc,argv);

    init_counter();

    fprintf(stderr,"Counting clusters separated by at leaset %d mismatch%s\n",dist+1,(dist+1==1)?"":"es");
    fprintf(stderr,"Using only sequences matching template \"%s\"\n",template);

    if (argv_index_of_first_path > 0) { int i; for (i=argv_index_of_first_path; i<argc; i++) { load_barcodes_from_file(argv[i]); } }
    else { load_barcodes_from_file(0); }

    if (!n_barcodes_total) { exit(0); }

    fprintf(stderr,"Loaded %d sequences\n",n_barcodes_total);
    /*fprintf(stderr,"Stored in %d blocks (occupying %" PRINT_SIZE_T " bytes in memory)\n",n_blocks,sizeof(block_t)*n_blocks);*/

    /*print_all_barcode_blocks();*/

    fprintf(stderr,"Sorting ..");
    store_barcodes_in_volumes();
    sort_barcodes();
    fprintf(stderr," OK\n");

    /*print_all_barcode_volumes();*/

    remove_duplicates();

    /*print_all_barcode_volumes();*/

    if (!dist)
    {
        printf("%lld clusters\n",n_unique_barcodes);
        exit(0);
    }

#if 0
    {
        unit_t u[2] = {291147936921752ull,16041821872693706752ull};
        /*unit_t w[2] = {16041821872693706752ull,291147936921752ull};*/
        unit_t w[2] = {0xFFFFFFFFFFFFFFFFull,0xFFFFFFFFFFFFFFFFull};
        int i;
        /*for (i=0; i<barcode_complexity; i++)
        {
            unsigned bn = read_barcode(u,i);
            fprintf(stderr,"%2d %2d %c\n",i,bn,binary_to_nucleotide[bn]);
        }*/
        for (i=0; i<barcode_complexity; i++)
        {
            unsigned bn = read_barcode(u,i);
            unsigned bn2;
            write_barcode(w,i,bn);
            bn2 = read_barcode(w,i);
            fprintf(stderr,"%2d  %d -> %d\n",i,bn,bn2);
        }
        fprintf(stderr,"%llu %llu\n",w[0],w[1]);
    }
#endif

    create_clusters();

    /*cluster_all_by_mutating();*/
    cluster_all_by_comparing();

    count_clusters();

    return 0;
}

#endif
